from numpy.testing import assert_allclose
import numpy as np
import os
from kawin.Thermodynamics import BinaryThermodynamics, MulticomponentThermodynamics
from kawin.Surrogate import BinarySurrogate, MulticomponentSurrogate
from kawin.tests.datasets import *

AlZrTherm = BinaryThermodynamics(ALZR_TDB, ['AL', 'ZR'], ['FCC_A1', 'AL3ZR'], drivingForceMethod='approximate')
NiCrAlTherm = MulticomponentThermodynamics(NICRAL_TDB, ['NI', 'CR', 'AL'], ['FCC_A1', 'FCC_L12'], drivingForceMethod='approximate')

#Set constant sampling densities for each Thermodynamics object
#pycalphad equilibrium results may change based off sampling density, so this is to make sure
#tests won't failed unneccesarily because the default sampling densities are modified
AlZrTherm.setDFSamplingDensity(2000)
AlZrTherm.setEQSamplingDensity(500)
NiCrAlTherm.setDFSamplingDensity(2000)
NiCrAlTherm.setEQSamplingDensity(500)

def test_Surr_binary_DG_output():
    '''
    Tests output of binary surrogate driving force function
    This should give the same response as corresponding functions
    in Thermodynamics

    Ex. for f(x, T) -> (dg, xP)
        (scalar, scalar) input -> scalar
        (array, array) input -> array
    '''
    surr = BinarySurrogate(AlZrTherm)
    T = 673.15
    xtrain = np.logspace(-5, -2, 5)
    surr.trainDrivingForce(xtrain, T, scale='log')

    dg, xP = surr.getDrivingForce(xtrain[3], 673.15, returnComp=True)
    dgT, xPT = AlZrTherm.getDrivingForce(xtrain[3], 673.15, returnComp=True, training = True)
    dgarray, xParray = surr.getDrivingForce([0.004, 0.005], [673.15, 683.15], returnComp=True)

    assert np.isscalar(dg) or (type(dg) == np.ndarray and dg.ndim == 0)
    assert np.isscalar(xP) or (type(xP) == np.ndarray and xP.ndim == 0)
    assert hasattr(dgarray, '__len__') and len(dgarray) == 2
    assert hasattr(xParray, '__len__') and len(xParray) == 2

    #Compare to Thermodynamics, high tolerance since we're just checking that functions are interchangeable
    assert_allclose(dg, dgT, rtol=1e-1)
    assert_allclose(xP, xPT, rtol=1e-1)

def test_Surr_binary_IC_output():
    '''
    Tests output of binary surrogate composition function
    This should give the same response as corresponding functions
    in Thermodynamics

    Ex. for f(T, g) -> (xM, xP)
        (scalar, scalar) -> (scalar, scalar)
        (array, array) -> (array, array)
        (scalar, array) -> (array, array)   Special case where T is scalar
    '''
    surr = BinarySurrogate(AlZrTherm)
    T = 673.15

    gExtra = np.linspace(100, 10000, 5)
    surr.trainInterfacialComposition(T, gExtra)

    xm, xp = surr.getInterfacialComposition(673.15, gExtra[3])
    xmT, xpT = AlZrTherm.getInterfacialComposition(673.15, gExtra[3])
    xmarray, xparray = surr.getInterfacialComposition([673.15, 683.15], [5000, 10000])
    xmarray2, xparray2 = surr.getInterfacialComposition(673.15, [5000, 10000])

    assert np.isscalar(xm) or (type(xm) == np.ndarray and xm.ndim == 0)
    assert np.isscalar(xp) or (type(xp) == np.ndarray and xp.ndim == 0)
    assert hasattr(xmarray, '__len__') and len(xmarray) == 2
    assert hasattr(xparray, '__len__') and len(xparray) == 2
    assert hasattr(xmarray2, '__len__') and len(xmarray2) == 2
    assert hasattr(xparray2, '__len__') and len(xparray2) == 2

    #Compare to Thermodynamics, high tolerance since we're just checking that functions are interchangeable
    assert_allclose(xm, xmT, rtol=1e-1)
    assert_allclose(xp, xpT, rtol=1e-1)


def test_Surr_binary_Diff_output():
    '''
    Tests output of binary surrogate diffusivity function
    This should give the same response as corresponding functions
    in Thermodynamics

    Ex. for diffusivity
        f(x, T) = diff
        (scalar, scalar) -> scalar
        (array, array) -> array
    '''
    surr = BinarySurrogate(AlZrTherm)
    T = 673.15
    xtrain = np.logspace(-5, -2, 5)

    surr.trainInterdiffusivity(xtrain, [T, T + 100])

    dnkj = surr.getInterdiffusivity(xtrain[3], 673.15)
    dnkjT = AlZrTherm.getInterdiffusivity(xtrain[3], 673.15)
    dnkjarray = surr.getInterdiffusivity([0.004, 0.005], [673.15, 683.15])
    assert np.isscalar(dnkj) or (type(dnkj) == np.ndarray and dnkj.ndim == 0)
    assert hasattr(dnkjarray, '__len__') and len(dnkjarray) == 2

    #Compare to Thermodynamics, high tolerance since we're just checking that functions are interchangeable
    assert_allclose(dnkj, dnkjT, rtol=1e-1)

def test_Surr_binary_save():
    '''
    Checks that binary surrogate can be saved and loaded to get same values
    '''
    surr = BinarySurrogate(AlZrTherm)
    T = 673.15
    xtrain = np.logspace(-5, -2, 5)
    surr.trainDrivingForce(xtrain, T, scale='log')

    gExtra = np.linspace(100, 1000, 5)
    surr.trainInterfacialComposition(T, gExtra)

    surr.trainInterdiffusivity(xtrain, [T, T + 100])

    a, b = surr.getDrivingForce(0.004, T, True)
    c, d = surr.getInterfacialComposition(T, 500)
    e = surr.getInterdiffusivity(0.1, T + 50)

    surr.save('kawin/tests/alzr')

    surr2 = BinarySurrogate.load('kawin/tests/alzr')
    a2, b2 = surr2.getDrivingForce(0.004, T, True)
    c2, d2 = surr2.getInterfacialComposition(T, 500)
    e2 = surr2.getInterdiffusivity(0.1, T + 50)

    os.remove('kawin/tests/alzr')

    assert_allclose([a, b, c, d, e], [a2, b2, c2, d2, e2], rtol=1e-3)

def test_Surr_binary_save_missing():
    '''
    Checks that load function will not fail if one of the three surrogates are not trained yet
    '''
    surr = BinarySurrogate(AlZrTherm)
    T = 673.15
    xtrain = np.logspace(-5, -2, 5)
    surr.trainDrivingForce(xtrain, T, scale='log')

    a, b = surr.getDrivingForce(0.004, T)

    surr.save('kawin/tests/alzr')

    surr2 = BinarySurrogate.load('kawin/tests/alzr')
    a2, b2 = surr2.getDrivingForce(0.004, T)

    os.remove('kawin/tests/alzr')

    assert_allclose(a, a2, rtol=1e-3)

def test_Surr_ternary_DG_output():
    '''
    Tests output of multicomponent surrogate driving force function
    This should give the same response as corresponding functions
    in Thermodynamics

    Ex. for f(x, T) -> (dg, xP)
        (array, scalar) -> scalar
        (2D array, array) -> array
    '''
    surr = MulticomponentSurrogate(NiCrAlTherm)
    T = [1073.15, 1123.15]
    x = [[0.06, 0.08], [0.06, 0.1], [0.06, 0.12], [0.08, 0.08], [0.08, 0.1], [0.08, 0.12], [0.1, 0.08], [0.1, 0.1], [0.1, 0.12]]
    surr.trainDrivingForce(x, T)

    dg, xP = surr.getDrivingForce(x[5], 1073.15, returnComp=True)
    dgT, xPT = NiCrAlTherm.getDrivingForce(x[5], 1073.15, returnComp=True, training = True)
    dgarray, xParray = surr.getDrivingForce([[0.08, 0.1], [0.085, 0.1], [0.09, 0.1]], [1073.15, 1078.15, 1083.15], returnComp=True)

    assert np.isscalar(dg) or (type(dg) == np.ndarray and dg.ndim == 0)
    assert xP.ndim == 1 and len(xP) == 2
    assert hasattr(dgarray, '__len__')
    assert xParray.shape == (3, 2)

    #Compare to Thermodynamics, high tolerance since we're just checking that functions are interchangeable
    assert_allclose(dg, dgT, rtol=1e-1)
    assert_allclose(xP, xPT, rtol=1e-1)

def test_Surr_ternary_IC_output():
    '''
    Tests output of multicomponent surrogate interfacial composition function
    This should give the same response as corresponding functions
    in Thermodynamics

    Ex. f(x, T, dG, R, gE) -> (gr, xM, xP, xM_EQ, xP_EQ)
        (array, scalar, scalar, scalar, scalar) -> (scalar, array, array, array, array)
        (array, scalar, scalar, array, array) -> (array, 2D array, 2D array, array, array)
    '''
    surr = MulticomponentSurrogate(NiCrAlTherm)
    T = [1073.15, 1123.15]
    x = [[0.06, 0.08], [0.06, 0.1], [0.06, 0.12], [0.08, 0.08], [0.08, 0.1], [0.08, 0.12], [0.1, 0.08], [0.1, 0.1], [0.1, 0.12]]
    surr.trainCurvature(x, T)

    g, ca, cb, caEQ, cbEQ = surr.getGrowthAndInterfacialComposition(x[5], 1073.15, 900, 1e-9, 1000)
    gT, caT, cbT, _, _ = NiCrAlTherm.getGrowthAndInterfacialComposition(x[5], 1073.15, 900, 1e-9, 1000, training = True)
    garray, caarray, cbarray, caEQarray, cbEQarray = surr.getGrowthAndInterfacialComposition([0.08, 0.1], 1073.15, 900, [0.5e-9, 1e-9, 2e-9], [2000, 1000, 500])

    assert np.isscalar(g) or (type(g) == np.ndarray and g.ndim == 0)
    assert hasattr(ca, '__len__') and len(ca) == 2
    assert hasattr(cb, '__len__') and len(cb) == 2
    assert hasattr(caEQ, '__len__') and len(caEQ) == 2
    assert hasattr(cbEQ, '__len__') and len(cbEQ) == 2
    assert hasattr(garray, '__len__') and len(garray) == 3
    assert caarray.shape == (3, 2)
    assert cbarray.shape == (3, 2)
    assert hasattr(caEQarray, '__len__') and len(caEQarray) == 2
    assert hasattr(caEQarray, '__len__') and len(caEQarray) == 2

    #Compare to Thermodynamics, high tolerance since we're just checking that functions are interchangeable
    assert_allclose(g, gT, rtol=1e-1)
    assert_allclose(ca, caT, rtol=1e-1)
    assert_allclose(cb, cbT, rtol=1e-1)

def test_Surr_ternary_save():
    '''
    Checks that multicomponent surrogate can be saved and loaded
    '''
    surr = MulticomponentSurrogate(NiCrAlTherm)
    T = [1073.15, 1123.15]
    x = [[0.06, 0.08], [0.06, 0.1], [0.06, 0.12], [0.08, 0.08], [0.08, 0.1], [0.08, 0.12], [0.1, 0.08], [0.1, 0.1], [0.1, 0.12]]
    surr.trainDrivingForce(x, T)

    surr.trainCurvature(x, T)

    a, b = surr.getDrivingForce([0.08, 0.1], T[0]+25, True)
    g, ca, cb, _, _ = surr.getGrowthAndInterfacialComposition([0.08, 0.1], T[0]+25, 900, 1e-9, 1000)
    beta = surr.impingementFactor([0.08, 0.1], T[0]+25)

    surr.save('kawin/tests/nicral')

    surr2 = MulticomponentSurrogate.load('kawin/tests/nicral')
    a2, b2 = surr2.getDrivingForce([0.08, 0.1], T[0]+25, True)
    g2, ca2, cb2, _, _ = surr2.getGrowthAndInterfacialComposition([0.08, 0.1], T[0]+25, 900, 1e-9, 1000)
    beta2 = surr2.impingementFactor([0.08, 0.1], T[0]+25)

    os.remove('kawin/tests/nicral')

    assert_allclose([a, b[0], b[1], g, ca[0], ca[1], cb[0], cb[1], beta], [a2, b2[0], b2[1], g2, ca2[0], ca2[1], cb2[0], cb2[1], beta2], rtol=1e-3)

def test_Surr_ternary_save_missing():
    '''
    Checks that load function will not fail if one of the three surrogates are not trained yet
    '''
    surr = MulticomponentSurrogate(NiCrAlTherm)
    T = [1073.15, 1123.15]
    x = [[0.06, 0.08], [0.06, 0.1], [0.06, 0.12], [0.08, 0.08], [0.08, 0.1], [0.08, 0.12], [0.1, 0.08], [0.1, 0.1], [0.1, 0.12]]
    surr.trainDrivingForce(x, T)

    a, b = surr.getDrivingForce([0.08, 0.1], T[0]+25, True)
    surr.save('kawin/tests/nicral')

    surr2 = MulticomponentSurrogate.load('kawin/tests/nicral')
    a2, b2 = surr2.getDrivingForce([0.08, 0.1], T[0]+25, True)
    os.remove('kawin/tests/nicral')

    assert_allclose([a, b[0], b[1]], [a2, b2[0], b2[1]], rtol=1e-3)



import numpy as np

class ShapeFactor:
    '''
    Defines functions for shape factors of needle, plate and cuboidal precipitates
    Shape factors involve the following:
        Aspect Ratio - ratio of long axis to short axis
        Eccentricity - characteristic value of the precipitate shape that depends on the aspect ratio
        Equivalent radius factor - correction factor to the radius to give the radius of a sphere with equivalent volume
        Kinetic factor - correction factor for the growth rate of the particle
        Thermodynamic factor / Area factor - correction factor for the critical driving force/radius for nucleation
        
        For a sphere (or needle and plate precipitates with aspect ratio of 1):
            Eccentricity = 0
            Equivalent radius factor = 1
            Kinetic factor = 1
            Thermodynamic factor = 1

    NOTE:
    normalRadaii, eqRadiusFactor, kineticFactor and thermoFactor 
        take in R and output the corresponding factors

    _normalRadaiiEquation, __eqRadiusEquation, _kineticEquation and _thermoEquation
        take in aspect ratio and output the corresponding factors
    '''
    
    #Particle shape types
    SPHERE = 0
    NEEDLE = 1
    PLATE = 2
    CUBIC = 3

    def __init__(self):
        self.setSpherical()
        self.setAspectRatio(1)

        #This is the same for all equations
        self.eqRadiusFactorMin = 1
        self.tol = 1e-3
        
    def setAspectRatio(self, ar):
        '''
        Sets aspect ratio as a function of R
        
        Parameters
        ----------
        ar : float or function
            Aspect ratio
            If function, must take in radius and return aspect ratio
        '''
        if np.isscalar(ar):
            self._aspectRatioScalar = ar
            self.aspectRatio = self._scalarAspectRatioEquation
            self.findRcrit = self._findRcritScalar
        else:
            self.aspectRatio = ar
            self.findRcrit = self._findRcrit
            
    def _scalarAspectRatioEquation(self, R):
        '''
        Aspect ratio as a function of radius if the aspect ratio is set as a scalar
        '''
        if hasattr(R, '__len__'):
            return self._aspectRatioScalar * np.ones(len(R))
        else:
            return self._aspectRatioScalar
            
    def setSpherical(self, ar = 1):
        '''
        Sets factors for spherical precipitates
        '''
        self.particleType = self.SPHERE
        self._eqRadiusEquation = self._eqRadiusFactorSphere
        self._normalRadiiEquation = self._normalRadiiSphere
        self._kineticEquation = self._kineticFactorSphere
        self._thermoEquation = self._thermoFactorSphere
        self.setAspectRatio(1)
        
        self.kineticFactorMin = 1
        self.thermoFactorMin = 1
        
    def setNeedleShape(self, ar = 1):
        '''
        Factors for needle shaped precipitates
        '''
        self.particleType = self.NEEDLE
        self.setAspectRatio(ar)
        self._eqRadiusEquation = self._eqRadiusFactorNeedle
        self._normalRadiiEquation = self._normalRadiiNeedle
        self._kineticEquation = self._kineticFactorEquationNeedle
        self._thermoEquation = self._thermoFactorEquationNeedle
        
        self.kineticFactorMin = 1
        self.thermoFactorMin = 1
        
    def setPlateShape(self, ar = 1):
        '''
        Factors for plate shaped precipitates
        '''
        self.particleType = self.PLATE
        self.setAspectRatio(ar)
        self._eqRadiusEquation = self._eqRadiusFactorPlate
        self._normalRadiiEquation = self._normalRadiiPlate
        self._kineticEquation = self._kineticFactorEquationPlate
        self._thermoEquation = self._thermoFactorEquationPlate
        
        self.kineticFactorMin = 1
        self.thermoFactorMin = 1
        
    def setCuboidalShape(self, ar = 1):
        '''
        Factors for cuboidal shaped precipitates
        '''
        self.particleType = self.CUBIC
        self.setAspectRatio(ar)
        self._eqRadiusEquation = self._eqRadiusFactorCuboidal
        self._normalRadiiEquation = self._normalRadiiCuboidal
        self._kineticEquation = self._kineticFactorEquationCuboidal
        self._thermoEquation = self._thermoFactorEquationCuboidal
        
        self.kineticFactorMin = 1
        
        #Thermodynamic factor for cuboidal precipitates is not 1 when aspect ratio is 1
        self.thermoFactorMin = self._thermoFactorEquationCuboidal(1)

    def normalRadii(self, R):
        '''
        Radius along the 3 axis to give a spherical volume of 1

        Parameters
        ----------
        R : float or array
            Equivalent spherical radius

        Returns
        -------
        3 length array for each axis if R is scalar
        n x 3 array if R is an array
        '''
        ar = self.aspectRatio(R)

        #Keep aspect ratio above 1
        if hasattr(ar, '__len__'):
            ar[ar < 1] = 1
        else:
            ar = 1 if ar < 1 else ar
            
        return self._normalRadiiEquation(ar)

    def eqRadiusFactor(self, R):
        '''
        Equivalent spherical radius factor for generic precipitates

        Parameters
        ----------
        R : float or array
            Equivalent spherical radius

        Returns
        -------
        Eq. radius factor with same shape as R
        '''
        ar = self.aspectRatio(R)
        if hasattr(ar, '__len__'):
            factor = self.eqRadiusFactorMin * np.ones(len(ar))
            factor[ar > 1] = self._eqRadiusEquation(ar[ar > 1])
            return factor
        else:
            if ar <= 1:
                return self.eqRadiusFactorMin
            else:
                return self._eqRadiusEquation(ar)
        
    def kineticFactor(self, R):
        '''
        Kinetic factor for generic precipitates

        Parameters
        ----------
        R : float or array
            Equivalent spherical radius

        Returns
        -------
        Kinetic factor with same shape as R
        '''
        ar = self.aspectRatio(R)
        if hasattr(ar, '__len__'):
            factor = self.kineticFactorMin * np.ones(len(ar))
            factor[ar > 1] = self._kineticEquation(ar[ar > 1])
            return factor
        else:
            if ar <= 1:
                return self.kineticFactorMin
            else:
                return self._kineticEquation(ar)
                
    def thermoFactor(self, R):
        '''
        Thermodynamic factor for generic precipitates

        Parameters
        ----------
        R : float or array
            Equivalent spherical radius

        Returns
        -------
        Thermodynamic factor with same shape as R
        '''
        ar = self.aspectRatio(R)
        if hasattr(ar, '__len__'):
            factor = self.thermoFactorMin * np.ones(len(ar))
            factor[ar > 1] = self._thermoEquation(ar[ar > 1])
            return factor
        else:
            if ar <= 1:
                return self.thermoFactorMin
            else:
                return self._thermoEquation(ar)
                
    def eccentricity(self, ar):
        '''
        Eccentricity given the aspect ratio (for needle and plate shaped precipitates)
        '''
        return np.sqrt(1 - 1 / ar**2)
        
    # Equivalent spherical radius ----------------------------------------
    def _eqRadiusFactorSphere(self, ar):
        '''
        Equivalent radius for a sphere (returns 1)
        '''
        return 1

    def _eqRadiusFactorNeedle(self, ar):
        '''
        Equivalent radius for needle shaped precipitate
        '''
        return np.cbrt(ar)

    def _eqRadiusFactorPlate(self, ar):
        '''
        Equivalent radius for plate shaped precipitate
        '''
        return np.cbrt(ar**2)

    def _eqRadiusFactorCuboidal(self, ar):
        '''
        Equivalent radius for cuboidal shaped precipitate
        '''
        return np.cbrt(3 * ar / (4 * np.pi))

    # Normalized radaii --------------------------------------------------
    def _normalRadiiSphere(self, ar):
        '''
        Returns radius along the 3-axis for a volume of 1
        '''
        if hasattr(ar, '__len__'):
            return np.cbrt((3 / 4 * np.pi)) * np.ones((len(ar), 3))
        else:
            return np.cbrt((3 / (4 * np.pi))) * np.ones(3)

    def _normalRadiiNeedle(self, ar):
        '''
        Returns radius along the 3-axis for a volume of 1
        '''
        scale = np.cbrt(1 / ar)
        if hasattr(ar, '__len__'):
            return np.cbrt((3 / (4 * np.pi))) * np.array([scale, scale, scale * ar]).T
        else:
            return np.cbrt((3 / (4 * np.pi))) * np.array([scale, scale, scale * ar])

    def _normalRadiiPlate(self, ar):
        '''
        Returns radius along the 3-axis for a volume of 1
        '''
        scale = np.cbrt(1 / ar**2)
        if hasattr(ar, '__len__'):
            return np.cbrt((3 / (4 * np.pi))) * np.array([scale * ar, scale * ar, scale]).T
        else:
            return np.cbrt((3 / (4 * np.pi))) * np.array([scale * ar, scale * ar, scale])

    def _normalRadiiCuboidal(self, ar):
        '''
        Returns radius along the 3-axis for a volume of 1
        '''
        scale = np.cbrt(1 / ar)
        if hasattr(ar, '__len__'):
            return np.array([scale, scale, scale * ar]).T
        else:
            return np.array([scale, scale, scale * ar])

    # Kinetic factor -----------------------------------------------------        
    def _kineticFactorSphere(self, ar):
        '''
        Kinetic factor for a sphere (returns 1)
        '''
        return 1

    def _kineticFactorEquationNeedle(self, ar):
        '''
        Kinetic factor for needle shaped precipitate
        '''
        ecc = self.eccentricity(ar)
        return 2 * np.cbrt(ar**2) * ecc / (np.log(1 + ecc) - np.log(1 - ecc))

    def _kineticFactorEquationPlate(self, ar):
        '''
        Kinetic factor for plate shaped precipitate
        '''
        ecc = self.eccentricity(ar)
        return ecc * np.cbrt(ar) / (np.arccos(0) - np.arccos(ecc))

    def _kineticFactorEquationCuboidal(self, ar):
        '''
        Kinetic factor for cuboidal shaped precipitate
        '''
        return 0.1 * np.exp(-0.091 * (ar - 1)) + 1.736 * np.sqrt(ar**2 - 1) / (np.cbrt(ar) * np.log(2 * ar**2 + 2 * ar * np.sqrt(ar**2 - 1) - 1))

    # Thermodynamic factor -----------------------------------------------
    def _thermoFactorSphere(self, ar):
        '''
        Thermodynamic factor for a sphere (returns 1)
        '''
        return 1
        
    def _thermoFactorEquationNeedle(self, ar):
        '''
        Thermodynamic factor for needle shaped precipitate
        '''
        ecc = self.eccentricity(ar)
        return (1 / (2 * ar**(2/3))) * (1 + ar / ecc * np.arcsin(ecc))
        
    def _thermoFactorEquationPlate(self, ar):
        '''
        Thermodynamic factor for plate shaped precipitate
        '''
        ecc = self.eccentricity(ar)
        return (1 / (2 * ar**(4/3))) * (ar**2 + (1 / (2 * ecc)) * np.log((1 + ecc) / (1 - ecc)))
        
    def _thermoFactorEquationCuboidal(self, ar):
        '''
        Thermodynamic factor for cuboidal shaped precipitate
        '''
        return (2 * ar + 1) / (2 * np.pi) * (4 * np.pi / (3 * ar))**(2/3)
    
    def _findRcritScalar(self, RcritSphere, Rmax):
        '''
        Critical radius given a scalar aspect ratio
        '''
        return RcritSphere * self.thermoFactor(RcritSphere)
        #return RcritSphere * self.eqRadiusFactor(RcritSphere)
    
    def _findRcrit(self, RcritSphere, Rmax):
        '''
        Critical radius given aspect ratio as a function of radius
        Found by bisection method
        '''
        minR = RcritSphere
        maxR = Rmax
        mid = (minR + maxR) / 2
        
        #Objective function is R = R_sphere * thermoFactor(ar(R))
        #Or R / (R_sphere * thermoFactor(ar(R))) - 1 = 0, this requires that the radius is within tol percent of true value
        fMin = minR / (RcritSphere * self.thermoFactor(minR)) - 1
        fMax = maxR / (RcritSphere * self.thermoFactor(maxR)) - 1
        fMid = mid / (RcritSphere * self.thermoFactor(mid)) - 1

        #fMin = min / (RcritSphere * self.eqRadiusFactor(min)) - 1
        #fMax = max / (RcritSphere * self.eqRadiusFactor(max)) - 1
        #fMid = mid / (RcritSphere * self.eqRadiusFactor(mid)) - 1

        n = 0
        while np.abs(fMid) > self.tol:
            if fMin * fMid >= 0:
                minR = mid
                fMin = fMid
            else:
                maxR = mid
                fMax = fMid
                
            mid = (minR + maxR) / 2
            fMid = mid / (RcritSphere * self.thermoFactor(mid)) - 1
            #fMid = mid / (RcritSphere * self.eqRadiusFactor(mid)) - 1
            
            n += 1
            if n == 100:
                return RcritSphere
                
        return mid
        


import numpy as np
import pickle
from scipy.interpolate import Rbf
import scipy.spatial.distance as spd

def generateTrainingPoints(*arrays):
    '''
    Creates all combinations of inputted arrays
    Used for creating training points in composition space for
    MulticomponentSurrogate

    Parameters
    ----------
    arrays - arrays along each dimension
        Order of arrays should correspond to the order of elements when defining thermodynamic functions
    '''
    return np.array(np.meshgrid(*arrays)).T.reshape(-1, len(arrays))

def _filter_points(inputs, outputs, tol = 1e-3):
    '''
    Filter set of input points such that the closest distance is above tolerance
    This is to avoid non-positive definite training matrices when creating surrogate models

    Parameters
    ----------
    inputs : m x n matrix of floats
        Input points of m observations in n-dimensional space
    outputs : list of array of floats
        Output points, each array must be m x n where
            m is number of observations and n is dimensions of output
    tol : float
        Tolerance for distance between two input points

    Outputs
    -------
    (filtered inputs, filtered outputs)
    filtered inputs - array of input points
    filtered outputs - array of output points
    '''
    #Make distance matrix
    distance = spd.squareform(spd.pdist(inputs))

    #Indices to remove
    indices = np.where((distance > 0) & (distance <= tol))
    indices = np.unique(indices[0][indices[0] < indices[1]])

    newInputs = np.delete(inputs, indices, axis=0)
    newOutputs = []
    for i in range(len(outputs)):
        newOutputs.append(np.delete(outputs[i], indices, axis=0))

    return newInputs, newOutputs

class BinarySurrogate:
    '''
    Handles surrogate models for driving force, interfacial composition and
    diffusivity in a binary system
    
    Parameters
    ----------
    binaryThermodynamics - BinaryThermodynamics (optional)
        Driving force and interfacial composition will be taken from this if not explicitly defined
        If None, then drivingForce and interfacialComposition must be defined
        
    drivingForce - function (optional)
        Function will take in (composition, temperature) and return driving force,
            where a positive value means that precipitation is favorable
        
    interfacialComposition - function (optional)
        Function will take in (temperature, excess gibbs free energy) and
            return tuple (parent composition, precipitate composition) or (None, None) if precipitate is unstable

    diffusivity - function (optional)
        Function will take in (composition, temperature) and return diffusivity

    precPhase : str (optional)
        Precipitate phase to consider if binaryThermodynamics is defined

    Note: if binaryThermodynamics is not defined, then drivingForce and
        interfacial composition needs to be defined
    '''
    def __init__(self, binaryThermodynamics = None, drivingForce = None, interfacialComposition = None, diffusivity = None, precPhase = None):
        self.binTherm = binaryThermodynamics
        self.precPhase = precPhase
        
        #If no driving force or interfacial composition function is supplied, then use function from thermodynamics class
        if drivingForce is None:
            self.drivingForceFunction = lambda x, T, returnComp=False, training = True: self.binTherm.getDrivingForce(x, T, self.precPhase, returnComp, training)
        else:
            self.drivingForceFunction = drivingForce
        
        if interfacialComposition is None:
            self.interfacialCompositionFunction = lambda T, ge: self.binTherm.getInterfacialComposition(T, ge, self.precPhase)
        else:
            self.interfacialCompositionFunction = interfacialComposition

        if binaryThermodynamics is not None:
            self.diffusivityFunction = self.binTherm.getInterdiffusivity
        elif diffusivity is not None:
            self.diffusivityFunction = diffusivity

        self.eps = 1e-3

        #Driving force variables -----------------------------------------
        #Training data points
        self.drivingForce = []
        self.precComp = []
        self.dGcoords = []
        self.precCompIndices = []
        self.precCompCoords = []

        #Scaling factor to normalize distance
        self.XDGscale = None
        self.TDGscale = None

        #Surrogates
        self.DGfunction = 'linear'
        self.DGepsilon = 1
        self.DGsmooth = 0
        self.linearDG = True
        self.LogSurrDG = None
        self.LogSurrPrecComp = None
        self.SurrogateDrivingForce = None
        self.SurrogatePrecComp = None

        #Interfacial composition -----------------------------------------
        #Training data
        self.xParent = []
        self.xPrec = []
        self.ICcoords = []
        self.uniqueXPrec = []
        self.Gcoords = []
        self.G = []
        
        #Scaling factor to normalize distance
        self.TICscale = None
        self.GICscale = None
        self.XGscale = None
        
        #Surrogates
        self.ICfunction = 'linear'
        self.ICepsilon = 1
        self.ICsmooth = 0
        self.linearIC = True
        self.SurrogateParent = None
        self.SurrogatePrec = None
        self.SurrogateG = None
        self.LogSurrParent = None
        self.LogSurrPrec = None

        #Diffusivity -----------------------------------------------------
        #Training data
        self.Diffcoords = []
        self.Diff = []

        #Scaling factor
        #Scale temperature and free energy range is proportional to amount of data along given axis
        self.XDiffscale = None
        self.singleXDiff = False
        self.TDiffscale = None

        #Surrogates
        self.linearDiff = False
        self.Difffunction = 'linear'
        self.Diffepsilon = 1
        self.Diffsmooth = 0
        self.SurrogateDiff = None
        self.LogSurrDiff = None
        
        
    def trainDrivingForce(self, comps, temperature, function='linear', epsilon=1, smooth=0, scale='linear'):
        '''
        Creates training points and sets up surrogate models for driving force calculations
        
        Parameters
        ----------
        comps : array of floats
            Range of compositions for training points
        temperature : float or array
            Temperature or range of temperatures for training points
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the composition training data should be in log or linear scale
            Note: 'log' is recommended for dilute solutions
        '''
        #Convert temperature to array if not so
        if not hasattr(temperature, '__len__'):
            temperature = [temperature]
        
        #Ensure that composition are defined values if in log scale
        if scale == 'log':
            comps[comps == 0] = 1e-9
        
        self.drivingForce = []
        self.precComp = []
        self.dGcoords = []
        self.precCompIndices = []
        
        #Create training data
        n = 0   #Index for precCompIndices (needs to correspond to indices self.drivingForce array)
        for t in temperature:
            for x in comps:
                dG, xP = self.drivingForceFunction(x, t, returnComp = True)

                #If driving force can be obtained (generally True)
                if dG is not None:
                    self.drivingForce.append(dG)
                    self.dGcoords.append([x, t])
                    
                    #If driving force is positive, then store nucleate composition
                    if xP is not None:
                        self.precComp.append(xP)
                        self.precCompIndices.append(n)
                    
                    n += 1

        self.dGcoords = np.array(self.dGcoords)   
        self.drivingForce = np.array(self.drivingForce)
        self.precComp = np.array(self.precComp)
        self.precCompIndices = np.array(self.precCompIndices)
        
        #Log scale
        if scale == 'log':
            self.dGcoords[:,0] = np.log10(self.dGcoords[:,0])
        
        #Scale data so that it the range is proportional to the amount of data points along the given axis
        if len(comps) == 1:
            self.XDGscale = 1
        else:
            self.XDGscale = (np.amax(self.dGcoords[:,0]) - np.amin(self.dGcoords[:,0])) / len(comps)
            self.dGcoords[:,0] /= self.XDGscale
            
        if len(temperature) == 1:
            self.TDGscale = 1
        else:
            self.TDGscale = (np.amax(self.dGcoords[:,1]) - np.amin(self.dGcoords[:,1])) / len(temperature)
            self.dGcoords[:,1] /= self.TDGscale

        #Create new array of coordinates for precipitate composition (this is to allow for filtering)
        self.precCompCoords = self.dGcoords[self.precCompIndices]

        #Filter points such that all points are separated by a distance by at least self.eps
        self.dGcoords, outputs = _filter_points(self.dGcoords, [self.drivingForce], self.eps)
        self.drivingForce = outputs[0]
        
        self.precCompCoords, outputs = _filter_points(self.precCompCoords, [self.precComp], self.eps)
        self.precComp = outputs[0]

        if scale == 'log':
            self.linearDG = False
        else:
            self.linearDG = True
        self.DGfunction = function
        self.DGepsilon = epsilon
        self.DGsmooth = np.amax([smooth, self.eps])
        
        self._createDGSurrogate()

    def _createDGSurrogate(self):
        #Build surrogates
        if self.linearDG:
            self.SurrogateDrivingForce = Rbf(self.dGcoords[:,0], self.dGcoords[:,1], self.drivingForce, function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth)
            self.SurrogatePrecComp = Rbf(self.precCompCoords[:,0], self.precCompCoords[:,1], self.precComp, function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth)
        else:
            self.linearDG = False
            self.LogSurrDG = Rbf(self.dGcoords[:,0], self.dGcoords[:,1], self.drivingForce, function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth)
            self.LogSurrPrecComp = Rbf(self.precCompCoords[:,0], self.precCompCoords[:,1], self.precComp, function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth)
            self.SurrogateDrivingForce = lambda x, T: self.LogSurrDG(np.log10(x) / self.XDGscale, T)
            self.SurrogatePrecComp = lambda x, T: self.LogSurrPrecComp(np.log10(x) / self.XDGscale, T)

    def changeDrivingForceHyperparameters(self, function = 'linear', epsilon = 1, smooth = 0):
        '''
        Re-create surrogate model for driving force with updated hyperparameters

        Parameters
        ----------
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        '''
        if self.TDGscale is None:
            raise Exception("Driving force has not been trained.")

        self.DGfunction = function
        self.DGepsilon = epsilon
        self.DGsmooth = np.amax([smooth, self.eps])
        
        self._createDGSurrogate()            
        
    def getDrivingForce(self, x, T, returnComp = False):
        '''
        Gets driving force from surrogate models
        
        Parameters
        ----------
        x : float or array of floats
            Composition
        T : float or array of floats
            Temperature, must be same length as x
        returnComp : bool (optional)
            Returns precipitate composition if True (defaults to False)
        
        Returns
        -------
        (driving force, precipitate composition)
        Both will be same shape as x and T
        Positive driving force means that precipitate will form
        precipitate composition will be None if returnComp is False
        '''
        if self.TDGscale is None:
            raise Exception("Driving force has not been trained.")

        #Convert arrays to Numpy arrays for math operations
        if hasattr(x, '__len__'):
            x = np.array(x)
        if hasattr(T, '__len__'):
            T = np.array(T)

        if self.linearDG:
            dG = self.SurrogateDrivingForce(x / self.XDGscale, T / self.TDGscale)
            
            if returnComp:
                xP = self.SurrogatePrecComp(x / self.XDGscale, T / self.TDGscale)
                return dG, xP
            else:
                if hasattr(dG, '__len__'):
                    return dG, np.full(dG.shape, None)
                else:
                    return dG, None
                
        else:
            dG = self.SurrogateDrivingForce(x, T / self.TDGscale)
            
            if returnComp:
                xP = self.SurrogatePrecComp(x, T / self.TDGscale)

                return dG, xP
            else:
                if hasattr(dG, '__len__'):
                    return dG, np.full(dG.shape, None)
                else:
                    return dG, None
        
    def trainInterfacialComposition(self, temperature, freeEnergy, function='linear', epsilon=1, smooth=0, scale = 'linear'):
        '''
        Creates training points and sets up surrogate models for interfacial composition

        Parameters
        ----------
        temperature : float or array
            Temperature or range of temperatures for training points
        freeEnergy : array of floats
            range of free energy values from Gibbs-Thomson contribution
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the matrix composition output should be in log or linear scale
            Note: 'log' is recommended for dilute solutions
        '''
        #Make temperature an array if not so
        if not hasattr(temperature, '__len__'):
            temperature = [temperature]
        
        self.xParent = []
        self.xPrec = []
        self.ICcoords = []
        
        #Create training data
        for t in temperature:
            for g in freeEnergy:
                xM, xP = self.interfacialCompositionFunction(t, g)
                
                #If precipitate can be form at T,G, then add to training array
                if xM is not None and xM > 0:
                    self.xParent.append(xM)
                    self.xPrec.append(xP)
                    
                    self.ICcoords.append([t, g])
                    
        self._buildInterfacialCompositionModels(temperature, freeEnergy, function, epsilon, smooth, scale)
        
    def trainInterfacialCompositionFromDrivingForceData(self, function='linear', epsilon=1, smooth=0, scale='linear'):
        '''
        Converts driving force data ([x, T] -> G) to interfacial composition data ([T, G] -> x)
        This may lead to inaccuracies in the precipitate composition since driving force calculations are done by sampling the free energy curve

        Parameters
        ----------
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the matrix composition output should be in log or linear scale
            Note: 'log' is recommended for dilute solutions
        '''
        self.xPrec = [x for x in self.precComp]
        
        #Take driving force training data and convert to interfacial composition training data
        if self.linearDG:
            self.xParent = [self.dGcoords[i, 0] * self.XDGscale for i in self.precCompIndices]
        else:
            self.xParent = [10**self.dGcoords[i, 0] * self.XDGscale for i in self.precCompIndices]
            
        self.ICcoords = []
        for i in self.precCompIndices:
            self.ICcoords.append([self.dGcoords[i, 1] * self.TDGscale, self.drivingForce[i]])
        
        self._buildInterfacialCompositionModels(np.unique(self.dGcoords[:, 1]) * self.TDGscale, self.drivingForce, function, epsilon, smooth, scale)
                    
    def _buildInterfacialCompositionModels(self, temperature, freeEnergy, function, epsilon, smooth, scale):
        '''
        Builds interfacial composition model (this is separate to allow for both training from new data or driving force data)

        Parameters
        ----------
        temperature - range of temperatures
        freeEnergy - range of free energies
        function - radial basis function
        epsilon - scale factor
        smooth - smoothing factor
        scale - linear or log scale
        '''
        #Create new training points finding the driving force for precipitation at the precipitate composition
        self.uniqueXPrec = np.unique(self.xPrec)
        if np.amax(self.uniqueXPrec) - np.amin(self.uniqueXPrec) < 1e-4:
            self.uniqueXPrec = np.array([self.uniqueXPrec[0]])
        self.Gcoords = []
        self.G = []
        
        for t in temperature:
            for x in self.uniqueXPrec:
                #Driving force calcs can be interpreted as the maximum free energy that can be contributed by the Gibbs-Thomson effect
                dG, _ = self.drivingForceFunction(x, t, returnComp = False)

                #if driving force can be obtained (generally True)
                if dG is not None:
                    self.G.append(dG)
                    self.Gcoords.append([x, t])
                    
                    #Add these points to the composition surrogate training data as an endpoint (these are for the minimum particle radius)
                    self.ICcoords.append([t, dG])
                    self.xParent.append(x)
                    self.xPrec.append(x)
        
        self.G = np.array(self.G)
        self.Gcoords = np.array(self.Gcoords)
        
        self.xParent = np.array(self.xParent)
        self.xPrec = np.array(self.xPrec)
        self.ICcoords = np.array(self.ICcoords)
        
        #Invert free energy coordinates - this will make the spacing more consistent
        self.ICcoords[:,1] =  1 / self.ICcoords[:,1]
        
        #Scale temperature and free energy range is proportional to amount of data along given axis
        if len(temperature) == 1:
            self.TICscale = 1
        else:
            self.TICscale = (np.amax(self.ICcoords[:,0]) - np.amin(self.ICcoords[:,0])) / len(temperature)
            self.ICcoords[:,0] /= self.TICscale
            
        if len(freeEnergy) == 1:
            self.GICscale = 1
        else:
            self.GICscale = (np.amax(self.ICcoords[:,1]) - np.amin(self.ICcoords[:,1])) / len(freeEnergy)
            self.ICcoords[:,1] /= self.GICscale
            
        if len(self.uniqueXPrec) == 1:
            self.XGscale = 1
        else:
            self.XGscale = (np.amax(self.Gcoords[:,0]) - np.amin(self.Gcoords[:,0])) / len(self.uniqueXPrec)
            self.Gcoords[:,0] /= self.XGscale
            
        self.Gcoords[:,1] /= self.TICscale

        #Filter points such that all points are separated by a distance by at least self.eps
        self.ICcoords, outputs = _filter_points(self.ICcoords, [self.xParent, self.xPrec], self.eps)
        self.xParent = outputs[0]
        self.xPrec = outputs[1]
        
        self.Gcoords, outputs = _filter_points(self.Gcoords, [self.G], self.eps)
        self.G = outputs[0]

        if scale == 'log':
            self.linearIC = False
        else:
            self.linearIC = True
        self.ICfunction = function
        self.ICepsilon = epsilon
        self.ICsmooth = np.amax([smooth, self.eps])

        self._createICSurrogate()

    def _createICSurrogate(self):
        #Build surrogates
        if self.linearIC:
            self.SurrogateParent = Rbf(self.ICcoords[:,0], self.ICcoords[:,1], self.xParent, function = self.ICfunction, epsilon = self.ICepsilon, smooth=self.ICsmooth)
            self.SurrogatePrec = Rbf(self.ICcoords[:,0], self.ICcoords[:,1], self.xPrec, function = self.ICfunction, epsilon = self.ICepsilon, smooth=self.ICsmooth)  
        else:
            self.LogSurrParent = Rbf(self.ICcoords[:,0], self.ICcoords[:,1], np.log10(self.xParent), function = self.ICfunction, epsilon = self.ICepsilon, smooth=self.ICsmooth)
            self.LogSurrPrec = Rbf(self.ICcoords[:,0], self.ICcoords[:,1], np.log10(self.xPrec), function = self.ICfunction, epsilon = self.ICepsilon, smooth=self.ICsmooth) 
            
            self.SurrogateParent = lambda T, gExtraInverse: 10**(self.LogSurrParent(T, gExtraInverse))
            self.SurrogatePrec = lambda T, gExtraInverse: 10**(self.LogSurrPrec(T, gExtraInverse))
           
        if len(self.G) == 1:
            self.SurrogateG = lambda x, T: self.G[0]
        else:
            self.SurrogateG = Rbf(self.Gcoords[:,0], self.Gcoords[:,1], self.G, function = 'linear', epsilon = 1)

    def changeInterfacialCompositionHyperparameters(self, function = 'linear', epsilon = 1, smooth = 0):
        '''
        Re-create surrogate model for interfacial composition with updated hyperparameters

        Parameters
        ----------
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        '''
        if self.TICscale is None:
            raise Exception("Interfacial composition has not been trained.")

        self.ICfunction = function
        self.ICepsilon = epsilon
        self.ICsmooth = np.amax([smooth, self.eps])

        self._createICSurrogate()
            
    def getInterfacialComposition(self, T, gExtra = 0):
        '''
        Gets Interfacial composition from surrogate models
        
        Parameters
        ----------
        T : float or array of floats
            Temperature
        gExtra : float or array of floats
            Free energy from Gibbs-Thomson contribution (must be same length as T)
        
        Returns
        -------
        (composition of parent phase, composition of precipitate phase)
        Composition will be in same shape as T and gExtra
        Will return (None, None) if gExtra is large enough that 
            precipitate becomes unstable
        '''
        if self.TICscale is None:
            raise Exception("Interfacial composition has not been trained.")

        #Convert arrays to Numpy arrays for math operations
        #Also raise gExtra to lowest training value to avoid erroneous values
        #   NOTE: Do not use this as a way to minimize the amount of training points!!
        #   While this is a safegaurd to avoid weird results, this creates a constant growth rate
        #   for any values of gExtra less than the lowest training point, which can lead to non-realistic
        #   values in the precipitate simulation (i.e. precipitates growing in an unsaturated matrix).
        #   Just train more points at lower gExtra values (larger radius sizes)
        if hasattr(gExtra, '__len__'):
            gExtra = np.array(gExtra)
            gExtra[1 / (gExtra * self.GICscale) > np.amax(self.ICcoords[:,1])] = 1 / (np.amax(self.ICcoords[:,1]) * self.GICscale)
        else:
            if (1 / (gExtra * self.GICscale)) > np.amax(self.ICcoords[:,1]):
                gExtra = 1 / (np.amax(self.ICcoords[:,1]) * self.GICscale)
        if hasattr(T, '__len__'):
            T = np.array(T)

        #If gExtra is array and T isn't, then convert T to array
        #This is to keep consistent with Thermodynamics counterpart
        if hasattr(gExtra, '__len__') and not hasattr(T, '__len__'):
            T = T * np.ones(len(gExtra))

        xM, xP = self.SurrogateParent(T / self.TICscale, 1 / (gExtra * self.GICscale)), self.SurrogatePrec(T / self.TICscale, 1 / (gExtra * self.GICscale))
        
        dG = self.SurrogateG(xP / self.XGscale, T / self.TICscale)
        if hasattr(xM, '__len__'):
            noneVals = (dG < gExtra)
            xM[noneVals] = None
            xP[noneVals] = None
        else:
            if dG < gExtra:
                xM = None
                xP = None
                    
        return xM, xP

    def trainInterdiffusivity(self, comps, temperature, function='linear', epsilon=1, smooth=0, scale='linear'):
        '''
        Trains interdiffusivity from mobility parameters

        Parameters
        ----------
        comps : array of floats
            Range of compositions for training points
        temperature : float or array
            Temperature or range of temperatures for training points
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the diffusivity output should be in log or linear scale
        '''
        #Convert composition and temperature to array if not so
        if not hasattr(comps, '__len__'):
            comps = [comps]

        if not hasattr(temperature, '__len__'):
            temperature = [temperature]

        self.Diffcoords = []
        self.Diff = []

        for x in comps:
            for t in temperature:
                self.Diff.append(self.diffusivityFunction(x, t))
                self.Diffcoords.append([x, t])

        self.Diffcoords = np.array(self.Diffcoords)
        self.Diff = np.array(self.Diff)

        #Scale temperature and free energy range is proportional to amount of data along given axis
        if len(comps) == 1:
            self.XDiffscale = 1
            self.singleXDiff = True
        else:
            self.XDiffscale = (np.amax(self.Diffcoords[:,0]) - np.amin(self.Diffcoords[:,0])) / len(comps)
            self.Diffcoords[:,0] /= self.XDiffscale
            self.singleXDiff = False

        if len(temperature) == 1:
            self.TDiffscale = 1
        else:
            self.TDiffscale = (np.amax(self.Diffcoords[:,1]) - np.amin(self.Diffcoords[:,1])) / len(temperature)
            self.Diffcoords[:,1] /= self.TDiffscale

        #Filter diffusivity points
        self.Diffcoords, outputs = _filter_points(self.Diffcoords, [self.Diff], self.eps)
        self.Diff = outputs[0]

        if scale == 'log':
            self.linearDiff = False
        else:
            self.linearDiff = True
        self.Difffunction = function
        self.Diffepsilon = epsilon
        self.Diffsmooth = np.amax([smooth, self.eps])

        self._createDiffSurrogate()

    def _createDiffSurrogate(self):
        #If only 1 data point, then create constant function
        if len(self.Diff) == 1:
            self.SurrogateDiff = lambda x, T: self.Diff[0]
            return

        #Build surrogates
        if self.linearDiff:
            if self.singleXDiff:
                self.SurrogateDiff = Rbf(self.Diffcoords[:,1], self.Diff, function = self.Difffunction, epsilon = self.Diffepsilon, smooth = self.Diffsmooth)
            else:
                self.SurrogateDiff = Rbf(self.Diffcoords[:,0], self.Diffcoords[:,1], self.Diff, function = self.Difffunction, epsilon = self.Diffepsilon, smooth = self.Diffsmooth) 
        else:
            if self.singleXDiff:
                self.LogSurrDiff = Rbf(self.Diffcoords[:,1], np.log10(self.Diff), function = self.Difffunction, epsilon = self.Diffepsilon, smooth = self.Diffsmooth)
                self.SurrogateDiff = lambda T: 10**(self.LogSurrDiff(T))
            else:
                self.LogSurrDiff = Rbf(self.Diffcoords[:,0], self.Diffcoords[:,1], np.log10(self.Diff), function = self.Difffunction, epsilon = self.Diffepsilon, smooth = self.Diffsmooth) 
                self.SurrogateDiff = lambda x, T: 10**(self.LogSurrDiff(x, T))

    def changeDiffusivityHyperparameters(self, function = 'linear', epsilon = 1, smooth = 0):
        '''
        Re-create surrogate model for diffusivity with updated hyperparameters

        Parameters
        ----------
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        '''
        if self.XDiffscale is None:
            raise Exception("Diffusivity has not been trained.")

        self.Difffunction = function
        self.Diffepsilon = epsilon
        self.Diffsmooth = np.amax([smooth, self.eps])

        self._createDiffSurrogate()

    def getInterdiffusivity(self, x, T):
        '''
        Returns interdiffusivity

        Parameters
        ----------
        x : float or array of floats
            Composition
        T : float or array of floats
            Temperature (must be same length as x)
        
        Returns
        -------
        diffusivity (same shape as x and T)
        '''
        #Convert arrays to numpy arrays for math operations
        if hasattr(x, '__len__'):
            x = np.array(x)
        if hasattr(T, '__len__'):
            T = np.array(T)

        if self.XDiffscale is None:
            raise Exception("Diffusivity has not been trained.")

        if self.singleXDiff:
            return self.SurrogateDiff(T / self.TDiffscale)
        else:
            return self.SurrogateDiff(x / self.XDiffscale, T / self.TDiffscale)
        
    def drivingForceTrainingTemperature(self):
        '''
        Returns the temperature coordinates of driving force training points
        '''
        return self.dGcoords[:,1] * self.TDGscale
        
    def drivingForceTrainingComposition(self):
        '''
        Returns the composition coordinates of driving force training points
        '''
        if self.linearDG:
            return self.dGcoords[:,0] * self.XDGscale
        else:
            return 10**(self.dGcoords[:,0] * self.XDGscale)
            
    def interfacialCompositionTrainingTemperature(self):
        '''
        Returns the temperature coordinates of interfacial composition training points
        '''
        return self.ICcoords[:,0] * self.TICscale
        
    def interfacialCompositionTrainingGibbsThomson(self):
        '''
        Returns the Gibbs-Thomson contribution coordinates of interfacial composition training points
        '''
        return 1 / (self.ICcoords[:,1] * self.GICscale)

    def save(self, fileName):
        '''
        Pickles surrogate data
        Note: this will remove the user-defined driving force and interfacial compositions
            This is not a problem; however, a loaded surrogate will not be
            able to be re-trained with different training points

        Parameters
        ----------
        fileName : str
        '''
        self.binTherm = None
        self.drivingForceFunction = None
        self.interfacialCompositionFunction = None
        self.diffusivityFunction = None

        self.LogSurrDG = None
        self.LogSurrPrecComp = None
        self.SurrogateDrivingForce = None
        self.SurrogatePrecComp = None

        self.SurrogateParent = None
        self.SurrogatePrec = None
        self.SurrogateG = None
        self.LogSurrParent = None
        self.LogSurrPrec = None

        self.SurrogateDiff = None
        self.LogSurrDiff = None

        with open(fileName, 'wb') as file:
            pickle.dump(self, file)

        #Re-create surrogates so that it could still be used after saving
        if self.XDGscale is not None:
            self._createDGSurrogate()
        if self.TICscale is not None:
            self._createICSurrogate()
        if self.XDiffscale is not None:
            self._createDiffSurrogate()

    def load(fileName):
        '''
        Loads data from a pickled surrogate and builds driving force and interfacial composition functions

        Parameters
        ----------
        fileName : str

        Returns
        -------
        BinarySurrogate object
        '''
        surr = None
        with open(fileName, 'rb') as file:
            surr = pickle.load(file)

        #Re-create surrogates so that it could still be used after saving
        if surr.XDGscale is not None:
            surr._createDGSurrogate()
        if surr.TICscale is not None:
            surr._createICSurrogate()
        if surr.XDiffscale is not None:
            surr._createDiffSurrogate()

        return surr
        
class MulticomponentSurrogate:
    '''
    Handles surrogate models for driving force, interfacial composition
        and growth rate in a multicomponent system
    
    Parameters
    ----------
    thermodynamics - MulticomponentThermodynamics (optional)
        Driving force, interfacial composition and 
            curvature functions will be taken from this
        If None, then drivingForce and curvature will need to be defined
        
    drivingForce - function (optional))
        Function will take in (composition, temperature) and return driving force
            where a positive value means that precipitation is favorable
            composition is an array for each element, excluding the reference element
        
    interfacialComposition - function (optional)
        Takes in (composition, temperature, gExtra) and returns matrix and precipitate composition
            composition is an array for each element, excluding the reference element
        Function should return (None, None) if precipitate is unstable

    curvature - function (optional)
        Function will take in (composition, temperature) and return the following:
            {D-1 dCbar / dCbar^T M-1 dCbar} - for calculating interfacial composition of matrix
            {1 / dCbar^T M-1 dCbar} - for calculating growth rate
            {Gb^-1 Ga} - for calculating precipitate composition
            Ca - interfacial composition of matrix phase
            Cb - interfacial composition of precipitate phase
        Function will return (None, None, None, None, None) if composition is outside two phase region
    
    precPhase : str (optional)
        Precipitate phase to consider if binaryThermodynamics is defined

    Note: if binaryThermodynamics is not defined, then drivingForce and
        interfacial composition needs to be defined
    '''
    def __init__(self, thermodynamics = None, drivingForce = None, interfacialComposition = None, curvature = None, precPhase = None):
        self.therm = thermodynamics
        self.precPhase = precPhase
        self.elements = self.therm.elements[1:-1]
        
        #Grab driving force and curvature function from thermodynamics class if not supplied
        if drivingForce is None:
            #self.drivingForceFunction = self.therm.getDrivingForce
            self.drivingForceFunction = lambda x, T, returnComp=False, training = True: self.therm.getDrivingForce(x, T, self.precPhase, returnComp, training)
            #self.drivingForceFunction = lambda x, T, returnComp=False: self.therm.drivingForceFromCurvature(x, T, self.precPhase, returnComp)
        else:
            self.drivingForceFunction = drivingForce

        if interfacialComposition is None:
            self.interfacialCompositionFunction = lambda x, T, gExtra: self.therm.getInterfacialComposition(x, T, gExtra, self.precPhase)
        else:
            self.interfacialCompositionFunction = interfacialComposition

        if curvature is None:
            #self.curvature = self.therm.curvatureFactor
            self.curvature = lambda x, T, training = True: self.therm.curvatureFactor(x, T, self.precPhase, training)
        else:
            self.curvature = curvature

        self.eps = 1e-3

        #Driving force ---------------------------------------------------
        #Training data
        self.drivingForce = []
        self.precComp = []
        self.dGcoords = []
        self.precCompIndices = []
        self.precCompCoords = []

        #Scaling factor
        self.XDGscale = []
        self.TDGscale = None
        
        #Surrogates
        self.linearDG = True
        self.DGfunction = 'linear'
        self.DGepsilon = 1
        self.DGsmooth = 0
        self.SurrogateDrivingForce = None
        self.SurrPrecComp = None
        self.SurrogatePrecComp = None
        self.LogSurrDG = None
        self.LogSurrPrecComp = None

        #Interfacial composition -----------------------------------------
        #Training data
        self.Dc = []
        self.Mc = []
        self.Gba = []
        self.beta = []
        self.Ca = []
        self.Cb = []
        self.ICcoords = []

        #Scaling factors
        self.XICscale = None
        self.TICscale = None

        #Interfacial composition surrogates
        self.linearIC = True
        self.ICfunction = 'linear'
        self.ICepsilon = 1
        self.ICsmooth = 0

        self.SurrDc = None
        self.SurrCa = None
        self.SurrCb = None
        self.SurrGba = None

        self.LogSurrDc = None
        self.LogSurrMc = None
        self.LogSurrBeta = None
        self.LogSurrCa = None
        self.LogSurrCb = None
        self.LogSurrGba = None
        
        self.SurrogateDc = None
        self.SurrogateMc = None
        self.SurrogateBeta = None
        self.SurrogateCa = None
        self.SurrogateCb = None
        self.SurrogateGba = None

        
    def trainDrivingForce(self, comps, temperature, function='linear', epsilon=1, smooth=0, scale='linear'):
        '''
        Creates training points and sets up surrogate models for driving force calculations
        
        Parameters
        ----------
        comps : 2-D array
            Range of compositions for training points
            0th axis represents an individual training point
            1st axis represents element composition
        temperature : float or array
            Range of temperatures for training points
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the composition training data should be in log or linear scale
            Note: 'log' is recommended for dilute solutions
        '''
        #Make temperature an array if not so
        if not hasattr(temperature, '__len__'):
            temperature = [temperature]

        #Ensure that composition are defined values when using log scale
        if scale == 'log':
            for x in comps:
                x[x == 0] = 1e-9
        
        #Create training data
        self.drivingForce = []
        self.precComp = []
        self.dGcoords = []
        self.precCompIndices = []
        
        n = 0   #Index for precCompIndices (needs to correspond to indices self.drivingForce array)
        for t in temperature:
            for x in comps:
                dG, xP = self.drivingForceFunction(x, t, returnComp = True)
                
                #If driving force can be obtained (generally True)
                if dG is not None:
                    self.drivingForce.append(dG)
                    self.dGcoords.append(np.concatenate((x, [t])))
                    
                    #If driving force is positive, then add nucleate composition
                    #Also determine equilibrium matrix composition, then add training point where driving force is 0
                    
                    if xP is not None:
                        self.precComp.append(xP)
                        self.precCompIndices.append(n)
                        
                        xMeq, xPeq = self.interfacialCompositionFunction(x, t, 0)
                        if xMeq is not None:
                            self.drivingForce.append(0)
                            self.dGcoords.append(np.concatenate((xMeq[1:], [t])))
                            n += 1
                            self.precComp.append(xPeq[1:])
                            self.precCompIndices.append(n)
                        
                    n += 1
                
        self.drivingForce = np.array(self.drivingForce)
        self.precComp = np.array(self.precComp)
        self.dGcoords = np.array(self.dGcoords)
        self.precCompIndices = np.array(self.precCompIndices)
        
        #Log scale on only composition (good for dilute solutions)
        if scale == 'log':
            self.dGcoords[:,:-1] = np.log10(self.dGcoords[:,:-1])
        
        #Scale data so range is proportional to amount of data along given axis
        if len(comps) == 1:
            self.XDGscale = np.ones(len(self.elements))
        else:
            self.XDGscale = (np.amax(self.dGcoords[:,:-1]) - np.amin(self.dGcoords[:,:-1])) / len(comps)
            self.dGcoords[:,:-1] /= self.XDGscale
            
        if len(temperature) == 1:
            self.TDGscale = 1
        else:
            self.TDGscale = (np.amax(self.dGcoords[:,-1]) - np.amin(self.dGcoords[:,-1])) / len(temperature)
            self.dGcoords[:,-1] /= self.TDGscale

        #Create new array of coordinates for precipitate composition (this is to allow for filtering)
        self.precCompCoords = self.dGcoords[self.precCompIndices]

        #Filter points such that all points are separated by a distance by at least self.eps
        self.dGcoords, outputs = _filter_points(self.dGcoords, [self.drivingForce], self.eps)
        self.drivingForce = outputs[0]
        
        self.precCompCoords, outputs = _filter_points(self.precCompCoords, [self.precComp], self.eps)
        self.precComp = outputs[0]

        if scale == 'log':
            self.linearDG = False
        else:
            self.linearDG = True
        self.DGfunction = function
        self.DGepsilon = epsilon
        self.DGsmooth = np.amax([smooth, self.eps])

        self._createDGSurrogate()

    def _createDGSurrogate(self):
        #Build surrogates
        if self.linearDG:
            arguments = [self.dGcoords[:,i] for i in range(len(self.dGcoords[0]))]
            self.SurrogateDrivingForce = Rbf(*arguments, self.drivingForce, function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth)
            
            arguments = [self.precCompCoords[:,i] for i in range(len(self.dGcoords[0]))]
            self.SurrPrecComp = [Rbf(*arguments, self.precComp[:,i], function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth) for i in range(len(self.elements))]
            self.SurrogatePrecComp = lambda x, T: np.array([self.SurrPrecComp[i](*x, T) for i in range(len(self.elements))])
        else:
            arguments = [self.dGcoords[:,i] for i in range(len(self.dGcoords[0]))]
            self.LogSurrDG = Rbf(*arguments, self.drivingForce, function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth)
            self.LogSurrPrecComp = [Rbf(*arguments, self.precComp[:,i], function=self.DGfunction, epsilon=self.DGepsilon, smooth=self.DGsmooth) for i in range(len(self.elements))]
            self.SurrogateDrivingForce = lambda x, T: self.LogSurrDG(*(np.log10(x) / self.XDGscale), T)
            self.SurrogatePrecComp = lambda x, T: np.array([self.LogSurrPrecComp[i](*(np.log10(x) / self.XDGscale), T) for i in range(len(self.elements))])

    def changeDrivingForceHyperparameters(self, function = 'linear', epsilon = 1, smooth = 0):
        '''
        Re-create surrogate model for driving force with updated hyperparameters

        Parameters
        ----------
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        '''
        if self.TDGscale is None:
            raise Exception("Driving force has not been trained.")

        self.DGfunction = function
        self.DGepsilon = epsilon
        self.DGsmooth = np.amax([smooth, self.eps])
        
        self._createDGSurrogate()
        
    def getDrivingForce(self, x, T, returnComp = False):
        '''
        Gets driving force from surrogate models
        
        Parameters
        ----------
        x : array or 2D array
            Composition (array of float for each minor element)
            2D arrays will have 0th axis for each set and 1st axis for composition
        T : float or array
            Temperature (must be float or same length as 0th axis of x if array)
        
        Returns
        -------
        driving force (positive value means that precipitate is stable)
        '''
        if self.TDGscale is None:
            raise Exception("Driving force has not been trained.")

        if hasattr(T, '__len__'):
            T = np.array(T)
        x = np.array(x)
        if x.ndim == 2:
            x = x.T

        if self.linearDG:
            dG = self.SurrogateDrivingForce(*(x / self.XDGscale), T / self.TDGscale)
            
            if returnComp:
                return dG, self.SurrogatePrecComp(x / self.XDGscale, T / self.TDGscale).T
            else:
                if hasattr(dG, '__len__'):
                    return dG, np.full(dG.shape, None)
                else:
                    return dG, None
            
        else:
            dG = self.SurrogateDrivingForce(x, T / self.TDGscale)
            
            if returnComp:
                return dG, self.SurrogatePrecComp(x, T / self.TDGscale)
            else:
                if hasattr(dG, '__len__'):
                    return dG, np.full(dG.shape, None)
                else:
                    return dG, None

    def trainCurvature(self, comps, temperature, function='linear', epsilon=1, smooth=0, scale='linear'):
        '''
        Trains for curvature factor (from Phillipes and Voorhees - 2013) as a function of composition and temperature

        Creates 5 surrogates
        {D-1 dCbar / dCbar^T M-1 dCbar} - for calculating interfacial composition of matrix
        {1 / dCbar^T M-1 dCbar} - for calculating growth rate
        {Gb^-1 Ga} - for calculating precipitate composition
        Ca - interfacial composition of matrix phase
        Cb - interfacial composition of precipitate phase


        Parameters
        ----------
        comps : 2D array of floats
            Range of compositions for training points
            0th axis represents a training point
            1st axis represents element compositions
        temperature : float or array
            Range of temperatures for training points
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the matrix composition output should be in log or linear scale
            Note: 'log' is recommended for dilute solutions
        '''
        #Make temperature an array if not so
        if not hasattr(temperature, '__len__'):
            temperature = [temperature]

        #Ensure that composition are defined values when using log scale
        if scale == 'log':
            for x in comps:
                x[x == 0] = 1e-9
        
        #Create training data
        self.Dc = []
        self.Mc = []
        self.Gba = []
        self.beta = []
        self.Ca = []
        self.Cb = []
        self.ICcoords = []
        
        for t in temperature:
            for x in comps:
                dc, mc, gba, beta, ca, cb = self.curvature(x, t)

                if dc is not None:
                    #Since Dc, Mc and Gba is constant for a given tie-line, add 3 training data points (at bulk compostion and phase boundaries)
                    #This should give more accurate values at very small or very large supersaturations without having to calculate a lot of training data
                    compCoords = [x, ca, cb]
                    for i in range(3):
                        self.Dc.append(dc)
                        self.Mc.append(mc)
                        self.Gba.append(gba)
                        self.beta.append(beta)
                        self.Ca.append(ca)
                        self.Cb.append(cb)
                        self.ICcoords.append(np.concatenate((compCoords[i], [t])))

        self.Dc = np.array(self.Dc)
        self.Mc = np.array(self.Mc)
        self.Gba = np.array(self.Gba)
        self.beta = np.array(self.beta)
        self.Ca = np.array(self.Ca)
        self.Cb = np.array(self.Cb)
        self.ICcoords = np.array(self.ICcoords)
        
        #Log scale only on compositions (good for low solubility)
        if scale == 'log':
            self.ICcoords[:,:-1] = np.log10(self.ICcoords[:,:-1])
        
        #Scale data so range is proportional to amount of data along given axis
        if len(comps) == 1:
            self.XICscale = np.ones(len(self.elements))
        else:
            self.XICscale = (np.amax(self.ICcoords[:,:-1]) - np.amin(self.ICcoords[:,:-1])) / len(comps)
            self.ICcoords[:,:-1] /= self.XICscale
            
        if len(temperature) == 1:
            self.TICscale = 1
        else:
            self.TICscale = (np.amax(self.ICcoords[:,-1]) - np.amin(self.ICcoords[:,-1])) / len(temperature)
            self.ICcoords[:,-1] /= self.TICscale

        #Filter points such that all points are separated by a distance by at least self.eps
        self.ICcoords, outputs = _filter_points(self.ICcoords, [self.Dc, self.Mc, self.Gba, self.beta, self.Ca, self.Cb], self.eps)
        self.Dc = outputs[0]
        self.Mc = outputs[1]
        self.Gba = outputs[2]
        self.beta = outputs[3]
        self.Ca = outputs[4]
        self.Cb = outputs[5]

        if scale == 'log':
            self.linearIC = False
        else:
            self.linearIC = True
        self.ICfunction = function
        self.ICepsilon = epsilon
        self.ICsmooth = np.amax([smooth, self.eps])

        self._createICSurrogate()

    def _createICSurrogate(self):
        #Build surrogates
        if self.linearIC:
            arguments = [self.ICcoords[:,i] for i in range(len(self.ICcoords[0]))]
            self.SurrogateMc = Rbf(*arguments, self.Mc, function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth)
            self.SurrogateBeta = Rbf(*arguments, self.beta, function=self.ICfunction, epislon=self.ICepsilon, smooth=self.ICsmooth)

            self.SurrDc = [Rbf(*arguments, self.Dc[:,i], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for i in range(len(self.elements))]
            self.SurrCa = [Rbf(*arguments, self.Ca[:,i], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for i in range(len(self.elements))]
            self.SurrCb = [Rbf(*arguments, self.Cb[:,i], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for i in range(len(self.elements))]
            self.SurrGba = [[Rbf(*arguments, self.Gba[:,i,j], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for j in range(len(self.elements))] for i in range(len(self.elements))]
            
            self.SurrogateDc = lambda x, T: np.array([self.SurrDc[i](*x, T) for i in range(len(self.elements))])
            self.SurrogateCa = lambda x, T: np.array([self.SurrCa[i](*x, T) for i in range(len(self.elements))])
            self.SurrogateCb = lambda x, T: np.array([self.SurrCb[i](*x, T) for i in range(len(self.elements))])
            self.SurrogateGba = lambda x, T: np.array([[self.SurrGba[i][j](*x, T) for j in range(len(self.elements))] for i in range(len(self.elements))])
            
        else:
            arguments = [self.dGcoords[:,i] for i in range(len(self.dGcoords[0]))]
            self.LogSurrDc = [Rbf(*arguments, self.Dc[:,i], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for i in range(len(self.elements))]
            self.LogSurrMc = Rbf(*arguments, self.Mc, function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth)
            self.LogSurrBeta = Rbf(*arguments, self.beta, function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth)
            self.LogSurrCa = [Rbf(*arguments, self.Ca[:,i], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for i in range(len(self.elements))]
            self.LogSurrCb = [Rbf(*arguments, self.Cb[:,i], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for i in range(len(self.elements))]
            self.LogSurrGba = [[Rbf(*arguments, self.Gba[:,i,j], function=self.ICfunction, epsilon=self.ICepsilon, smooth=self.ICsmooth) for j in range(len(self.elements))] for i in range(len(self.elements))]
            
            self.SurrogateDc = lambda x, T: np.array([self.LogSurrDc[i](*(np.log10(x) / self.XICscale), T) for i in range(len(self.elements))])
            self.SurrogateMc = lambda x, T: self.LogSurrMc(*(np.log10(x) / self.XICscale), T)
            self.SurrogateBeta = lambda x, T: self.LogSurrBeta(*(np.log10(x) / self.XICscale), T)
            self.SurrogateCa = lambda x, T: np.array([self.LogSurrCa[i](*(np.log10(x) / self.XICscale), T) for i in range(len(self.elements))])
            self.SurrogateCb = lambda x, T: np.array([self.LogSurrCb[i](*(np.log10(x) / self.XICscale), T) for i in range(len(self.elements))])
            self.SurrogateGba = lambda x, T: np.array([[self.LogSurrGba[i][j](*(np.log10(x) / self.XICscale), T) for j in range(len(self.elements))] for i in range(len(self.elements))])

    def changeCurvatureHyperparameters(self, function = 'linear', epsilon = 1, smooth = 0):
        '''
        Re-create surrogate model for curvature factors with updated hyperparameters

        Parameters
        ----------
        function : str (optional)
            Radial basis function to use (defaults to 'linear')
            Other functions are 'multiquadric', 'inverse_multiquadric',
                'gaussian', 'cubic', 'quintic' and 'thin_plate'
        epsilon : float
            Scale for radial basis function (defaults to 1)
            Training data will be scaled automatically
                such that optimal scale is around 1
        smooth : float
            Smoothness of interpolation, (defaults to 0, where interpolation will go through all points)
        scale : float
            Whether the composition training data should be in log or linear scale
            Note: 'log' is recommended for dilute solutions
        '''
        if self.TICscale is None:
            raise Exception("Curvature has not been trained.")

        self.ICfunction = function
        self.ICepsilon = epsilon
        self.ICsmooth = np.amax([smooth, self.eps])

        self._createICSurrogate()

    def getCurvature(self, x, T):
        '''
        Gets driving force from surrogate models
        
        Parameters
        ----------
        x : array or 2D array
            Composition (array of float for each minor element)
            2D arrays will have 0th axis for each set and 1st axis for composition
        T : float or array
            Temperature (must be float or same length as 0th axis of x if array)
        
        Returns
        -------
        Curvature factors
            {D-1 dCbar / dCbar^T M-1 dCbar} - for calculating interfacial composition of matrix
            {1 / dCbar^T M-1 dCbar} - for calculating growth rate
            {Gb^-1 Ga} - for calculating precipitate composition
            Ca - interfacial composition of matrix phase
            Cb - interfacial composition of precipitate phase
        Note: this function currently does not return (None, None, None, None, None)
            if precipitate is unstable
        '''
        if self.TICscale is None:
            raise Exception("Curvature has not been trained.")

        if self.linearIC:
            dc = self.SurrogateDc(x / self.XICscale, T / self.TICscale)
            mc = self.SurrogateMc(*(x / self.XICscale), T / self.TICscale)
            gba = self.SurrogateGba(x / self.XICscale, T / self.TICscale)
            beta = self.SurrogateBeta(*(x / self.XICscale), T / self.TICscale)
            #beta = self.SurrogateBeta(x, T / self.TICscale)
            ca = self.SurrogateCa(x / self.XICscale, T / self.TICscale)
            cb = self.SurrogateCb(x / self.XICscale, T / self.TICscale)
            
            return dc, mc, gba, beta, ca, cb
            
        else:
            dc = self.SurrogateDc(x, T / self.TICscale)
            mc = self.SurrogateMc(x, T / self.TICscale)
            gba = self.SurrogateGba(x, T / self.TICscale)
            beta = self.SurrogateBeta(x, T / self.TICscale)
            ca = self.SurrogateCa(x, T / self.TICscale)
            cb = self.SurrogateCb(x, T / self.TICscale)
            
            return dc, mc, gba, beta, ca, cb

    def getGrowthAndInterfacialComposition(self, x, T, dG, R, gExtra):
        '''
        Returns growth rate and interfacial compostion given Gibbs-Thomson contribution

        Parameters
        ----------
        x : array of floats
            Matrix composition
        T : float
            Temperature
        dG : float
            Driving force
        R : float or array
            Precipitate radius
        gExtra : float or array
            Gibbs-Thomson contribution corresponding to R
            Must be same shape as R
        
        Returns
        -------
        (growth rate, matrix composition, precipitate composition)
        Growth rate will be float or array depending on R
        matrix and precipitate composition will be 1D or 2D array depending on R
            1D array will be length of composition
            2D array will have 0th axis be length of R and 1st axis be length of composition
        '''
        if self.TICscale is None:
            raise Exception("Curvature needs to be trained to calculated interfacial composition.")

        if hasattr(R, '__len__'):
            R = np.array(R)
        if hasattr(gExtra, '__len__'):
            gExtra = np.array(gExtra)

        dc, mc, gba, beta, ca, cb = self.getCurvature(x, T)

        Rdiff = (dG - gExtra)

        gr = (mc / R) * Rdiff

        if hasattr(Rdiff, '__len__'):
            calpha = np.zeros((len(Rdiff), len(self.elements)))
            dca = np.zeros((len(Rdiff), len(self.elements)))
            cbeta = np.zeros((len(Rdiff), len(self.elements)))
            for i in range(len(self.elements)):
                calpha[:,i] = x[i] - dc[i] * Rdiff
                dca[:,i] = calpha[:,i] - ca[i]

            dcb = np.matmul(gba, dca.T)
            for i in range(len(self.elements)):
                cbeta[:,i] = cb[i] + dcb[i,:]

            calpha[calpha < 0] = 0
            cbeta[cbeta < 0] = 0

            return gr, calpha, cbeta, ca, cb
        else:
            calpha = x - dc * Rdiff
            cbeta = cb + np.matmul(gba, (calpha - ca)).flatten()

            calpha[calpha < 0] = 0
            cbeta[cbeta < 0] = 0

            return gr, calpha, cbeta, ca, cb

    def impingementFactor(self, x, T):
        '''
        Calculates impingement factor for nucleation rate calculations

        Parameters
        ----------
        x : array of floats
            Matrix composition
        T : float
            Temperature
        '''
        if self.TICscale is None:
            raise Exception("Curvature needs to be trained to calculated impingement factor.")
        #return self.SurrogateBeta(x, T / self.TICscale)
        if self.linearIC:
            return self.SurrogateBeta(*(x / self.XICscale), T / self.TICscale)
        else:
            return self.SurrogateBeta(x, T / self.TICscale)

    def save(self, fileName):
        '''
        Pickles surrogate data
        Note: this will remove the user-defined driving force and curvature functions
            This is not a problem; however, a loaded surrogate will not be
            able to be re-trained with different training points

        Parameters
        ----------
        fileName : str
        '''
        self.therm = None
        self.drivingForceFunction = None
        self.interfacialCompositionFunction = None
        self.curvature = None

        self.SurrogateDrivingForce = None
        self.SurrPrecComp = None
        self.SurrogatePrecComp = None
        self.LogSurrDG = None
        self.LogSurrPrecComp = None

        self.SurrDc = None
        self.SurrCa = None
        self.SurrCb = None
        self.SurrGba = None

        self.LogSurrDc = None
        self.LogSurrMc = None
        self.LogSurrBeta = None
        self.LogSurrCa = None
        self.LogSurrCb = None
        self.LogSurrGba = None
        
        self.SurrogateDc = None
        self.SurrogateMc = None
        self.SurrogateBeta = None
        self.SurrogateCa = None
        self.SurrogateCb = None
        self.SurrogateGba = None

        with open(fileName, 'wb') as file:
            pickle.dump(self, file)

        if self.TDGscale is not None:
            self._createDGSurrogate()
        if self.TICscale is not None:
            self._createICSurrogate()

    def load(fileName):
        '''
        Loads data from a pickled surrogate and builds driving force and curvature functions

        Parameters
        ----------
        fileName : str

        Returns
        -------
        MulticomponentSurrogate object
        '''
        surr = None
        with open(fileName, 'rb') as file:
            surr = pickle.load(file)

        if surr.TDGscale is not None:
            surr._createDGSurrogate()
        if surr.TICscale is not None:
            surr._createICSurrogate()

        return surr
        


from numpy.testing import assert_allclose
import numpy as np
from kawin.ShapeFactors import ShapeFactor
from kawin.ElasticFactors import StrainEnergy

Rsingle = 2
Rarray = np.linspace(1, 2, 10)

def test_SphericalOutput():
    '''
    Tests output of spherical shape factors given a single radius or list of radii
    Since these factors are constant, all factors should give a scaler value

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setSpherical()

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_NeedleOutput():
    '''
    Tests output of needle shape factors given a single radius or list of radii

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setNeedleShape(ar=2)

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_PlateOutput():
    '''
    Tests output of plate shape factors given a single radius or list of radii

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setPlateShape(ar=2)

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_CuboidalOutput():
    '''
    Tests output of cuboidal shape factors given a single radius or list of radii

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setCuboidalShape(ar=2)

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_AspectRatio():
    '''
    Tests that aspect ratio can be as a function
    '''
    arFunc = lambda R : 0.5 * R + 3
    sf = ShapeFactor()
    sf.setAspectRatio(arFunc)
    Rarray = sf.aspectRatio(np.array([1, 2]))

    assert_allclose(Rarray, np.array([3.5, 4]), rtol=1e-5)

def test_StrainOutput():
    '''
    Tests that strain energy outputs:
        a) a scalar for a single set of radii (3 length array)
        b) a list for a list of radii (n x 3 array)
    '''
    se = StrainEnergy()
    se.setEigenstrain(0.01)
    se.setElasticConstants(150e9, 100e9, 75e9)

    rSingle = np.random.random(3)
    rArray = np.random.random((10, 3))

    elSingle = se.strainEnergy(rSingle)
    elArray = se.strainEnergy(rArray)

    assert np.isscalar(elSingle) or (type(elSingle) == np.ndarray and elSingle.ndim == 0)
    assert elArray.shape == (10,)


import numpy as np
import scipy.stats as sts
import matplotlib.pyplot as plt
import copy

class PopulationBalanceModel:
    '''
    Class for handling particle size distributions (PSD)
        This include time evolution, moments and graphing

    Parameters
    ----------
    cMin : float (optional)
        Lower bound of PSD (defaults at 0)
    cMax : float (optional)
        Upper bound of PSD (defaults at 100)
    bins : int (optional)
        Number of bins (defaults at 100)
    linearSpacing : bool (optional)
        If True (default), will store PSD on a linear scale
        Else, will store PSD on a logarithmic scale
    '''
    def __init__(self, cMin = 0, cMax = 100, bins = 100, linearSpacing = True):
        self.originalBins = bins
        self.bins = bins
        self.min = cMin
        self.max = cMax
        self.linear = linearSpacing
        
        self.reset()

        #Hidden variable for use in KWNEuler when determining composition assuming no diffusion in precipitate
        #Represents d(PSD)/dr * growth rate * dt
        #I would like this variable to be in KWNEuler, but this way is much easier
        self._fv = np.zeros(self.bins + 1)

        #Hidden variable for use in KWNEuler when adaptive time stepping is enabled
        #This allows for PSD to revert to its previous value if a time constraint is not met
        self._prevPSD = np.zeros(self.bins)
        
    def reset(self):
        '''
        Resets the PSD to 0
        This will remove any size classes that were added since initialization
        '''
        self.bins = self.originalBins
        if self.linear:
            self.PSDbounds = np.linspace(self.min, self.max, self.bins+1)
            self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
        else:
            self.PSDbounds = np.logspace(self.min, self.max, self.bins+1)
            self.PSDsize = np.power(10, 0.5 * (np.log(self.PSDbounds[1:]) + np.log(self.PSDbounds[:-1])))
            
        self.PSD = np.zeros(self.bins)

    def revert(self):
        self.PSD = copy.copy(self._prevPSD)

    def changeSizeClasses(self, cMin, cMax, bins, linearSpacing = True, resetPSD = True):
        '''
        Changes the size classes and resets the PSD
        
        Parameters
        ----------
        min : float
            Lower bound of PSD
        cMax : float
            Upper bound of PSD
        bins : int
            Number of bins
        linearSpacing : bool (optional)
            If True (default), will store PSD on a linear scale
            Else, will store PSD on a logarithmic scale
        resetPSD : bool (optional)
            Whether to reset the PSD (defaults to True)
        '''
        self.bins = bins
        self.min = cMin
        self.max = cMax
        self.linear = linearSpacing
        
        if linearSpacing:
            self.PSDbounds = np.linspace(cMin, cMax, bins+1)
            self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
        else:
            self.PSDbounds = np.logspace(cMin, cMax, bins+1)
            self.PSDsize = np.power(10, 0.5 * (np.log(self.PSDbounds[1:] * self.PSDbounds[:-1])))
            
        if resetPSD:
            self.reset()

    def addSizeClass(self):
        '''
        Adds an additional size class to end of distribution
        '''
        self.bins += 1
        self.PSDbounds = np.append(self.PSDbounds, 0)
        self.PSDsize = np.append(self.PSDsize, 0)
        self.PSD = np.append(self.PSD, 0)

        if self.linear:
            self.max += (self.PSDbounds[1] - self.PSDbounds[0])
            self.PSDbounds[-1] = self.max
            self.PSDsize[-1] = 0.5 * (self.PSDbounds[-1] + self.PSDbounds[-2])
        else:
            self.max = np.power(10, np.log(self.max * self.PSDbounds[1] / self.PSDbounds[0]))
            self.PSDbounds[-1] = self.max
            self.PSDsize[-1] = np.power(10, 0.5 * np.log(self.PSDbounds[-1] * self.PSDbounds[-2]))
        
    def LoadDistribution(self, data):
        '''
        Creates a particle size distribution from a set of data
        
        Parameters
        ----------
        data : array of floats
            Array of data to be inserted into PSD
        '''
        self.PSD, self.PSDbounds = np.histogram(data, self.PSDbounds)
        self.PSD = self.PSD.astype('float')
        
    def Normalize(self):
        '''
        Normalizes the PSD to 1
        '''
        sum = self.WeightedMoment(0, self.PSDbounds[1:] - self.PSDbounds[:-1])
        self.PSD /= sum
            
    def NormalizeToMoment(self, order = 0):
        '''
        Normalizes the PSD with so that the moment of specified order will be 1
        
        Parameters
        ----------
        order : int (optional)
            Order of moment that PSD will be normalized to (defaults to 0)
            Using zeroth order will normalize the PSD so the sum of PSD will be 1
        '''
        sum = self.Moment(order)
        self.PSD /= sum
        
    def Update(self, dt, flux, externalRates = None):
        '''
        Updates PSD given the flux and any external contributions

        Change in the amount of particles in a given size class = d(G*n)/dx
        Where G is flux of size class, n is number of particles in size class and dx is range of size class
        
        Parameters
        ----------
        dt : float
            Time increment
        flux : array
            Growth rate of each particle size class
            Array size must be (bins + 1) since this operates on bounds of size classes
        externalRates : array (optional)
            Additional contribution of particles (nucleation or dissolution)
            Array size must be (bins) since this operates on the individual size classes
        '''
        #Store current PSD
        self._prevPSD = copy.copy(self.PSD)

        netFlux = np.zeros(self.bins + 1)
        netFlux[0] = 0 if flux[0] > 0 else flux[0] * dt * self.PSD[0] / (self.PSDbounds[1] - self.PSDbounds[0])
        netFlux[-1] = 0 if flux[-1] < 0 else flux[-1] * dt * self.PSD[-1] / (self.PSDbounds[-1] - self.PSDbounds[-2])

        #If flux is going from size class n to n-1, then use size class n (flux <= 0)
        indices = flux[1:-1] <= 0
        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[1:][indices] / (self.PSDbounds[2:] - self.PSDbounds[1:-1])[indices]
        under = (netFlux[1:-1] < -self.PSD[1:]) & indices
        netFlux[1:-1][under] = -self.PSD[1:][under]
        
        #if flux is going from size class n-1 to n, then use size class n-1 (flux > 0)
        indices = ~indices
        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[:-1][indices] / (self.PSDbounds[1:-1] - self.PSDbounds[:-2])[indices]
        over = (netFlux[1:-1] > self.PSD[:-1]) & indices
        netFlux[1:-1][over] = self.PSD[:-1][over]

        self._fv = netFlux
        
        self.PSD += netFlux[:-1] - netFlux[1:]
        if externalRates is not None:
            self.PSD += externalRates
            
        #Set negative frequencies to 0
        self.PSD[self.PSD < 1] = 0

    def UpdateMomentConserved(self, dt, flux, externalRates = None, order = 1):
        '''
        Updates PSD given the flux and any external contributions

        Change in the amount of particles in a given size class = d(G*n^m)/dx
        Where G is flux of size class, n is number of particles in size class, m is moment order and dx is range of size class
        
        Parameters
        ----------
        dt : float
            Time increment
        flux : array
            Growth rate of each particle size class
            Array size must be (bins + 1) since this operates on bounds of size classes
        externalRates : array (optional)
            Additional contribution of particles (nucleation or dissolution)
            Array size must be (bins) since this operates on the individual size classes
        order : int (optional)
            Order of moment to conserve (defaults to 1, which is the same as the Update method)
        '''
        #Store current PSD
        self._prevPSD = copy.copy(self.PSD)
        
        netFlux = np.zeros(self.bins + 1)
        netFlux[0] = 0 if flux[0] > 0 else flux[0] * dt * self.PSD[0] * self.PSDsize[0]**order / (self.PSDbounds[1] - self.PSDbounds[0])
        netFlux[-1] = 0 if flux[-1] < 0 else flux[-1] * dt * self.PSD[-1] * self.PSDsize[-1]**order / (self.PSDbounds[-1] - self.PSDbounds[-2])

        #Note - this is not tested yet
        #If flux is going from size class n to n-1, then use size class n (flux <= 0)
        indices = flux[1:-1] <= 0
        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[1:][indices] * self.PSDsize[1:][indices]**order / (self.PSDbounds[2:] - self.PSDbounds[1:-1])[indices]
        under = (netFlux[1:-1] < -self.PSD[1:] * self.PSDsize[1:]**order) & indices
        netFlux[1:-1][under] = -self.PSD[1:][under] * self.PSDsize[1:][under]**order
        
        #if flux is going from size class n-1 to n, then use size class n-1 (flux > 0)
        indices = ~indices
        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[:-1][indices] * self.PSDsize[:-1][indices]**order / (self.PSDbounds[1:-1] - self.PSDbounds[:-2])[indices]
        over = (netFlux[1:-1] > self.PSD[:-1] * self.PSDsize[:-1]**order) & indices
        netFlux[1:-1][over] = self.PSD[:-1][over] * self.PSDsize[1:][over]**order

        self._fv = netFlux
        
        self.PSD += (netFlux[:-1] - netFlux[1:]) / self.PSDsize**order
        if externalRates is not None:
            self.PSD += externalRates
            
        #Set negative frequencies to 0
        self.PSD[self.PSD < 0] = 0
        
    def Moment(self, order):
        '''
        Moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        '''
        return np.sum(self.PSD * self.PSDsize**order)

    def CumulativeMoment(self, order):
        '''
        Cumulative distribution using moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        '''
        return np.cumsum(self.PSD * self.PSDsize**order)
        
    def WeightedMoment(self, order, weights):
        '''
        Weighted moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        weights : array
            Weights to apply to each size class
            Array size of (bins)
        '''
        return np.sum(self.PSD * self.PSDsize**order * weights)

    def CumulativeWeightedMoment(self, order, weights):
        '''
        Weighted moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        weights : array
            Weights to apply to each size class
            Array size of (bins)
        '''
        return np.cumsum(self.PSD * self.PSDsize**order * weights)

    def ZeroMoment(self):
        '''
        Sum of the PSD
        '''
        return np.sum(self.PSD)
    
    def FirstMoment(self):
        '''
        Length weighted moment
        '''
        return np.sum(self.PSD * self.PSDsize)
        
    def SecondMoment(self):
        '''
        Area weighted moment
        '''
        return np.sum(self.PSD * self.PSDsize**2)
        
    def ThirdMoment(self):
        '''
        Volume weighted moment
        '''
        return np.sum(self.PSD * self.PSDsize**3)
        
    def PlotCurve(self, axes, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the PSD as a curve
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        if fill:
            axes.fill_between(self.PSDsize * scale, self.PSD, np.zeros(len(self.PSD)), *args, **kwargs)
        else:
            axes.plot(self.PSDsize * scale, self.PSD, *args, **kwargs)
        self.setAxes(axes, logX, logY) 

    def PlotDistributionDensity(self, axes, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the distribution density as a curve
        Defined as N_i / (R_i+1/2 - R_i-1/2)
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        if fill:
            axes.fill_between(self.PSDsize * scale, self.PSD, np.zeros(len(self.PSD)), *args, **kwargs)
        else:
            axes.plot(self.PSDsize * scale, self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1]), *args, **kwargs)
        
        #Set x-limits
        if logX:
            if self.min == 0:
                axes.set_xlim([self.PSDbounds[1], self.max])
            else:
                axes.set_xlim([self.min, self.max])
            axes.set_xscale('log')
        else:
            axes.set_xlim([self.min, self.max]) 

        #Set y-limits
        if logY:
            axes.set_ylim([1e-1, np.amax([1.1 * np.max(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])
            axes.set_yscale('log')
        else:
            axes.set_ylim([0, np.amax([1.1 * np.max(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])

    def PlotKDE(self, axes, bw_method = None, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the kernel density estimation (KDE)
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        bw_method : str (optional)
            Method to estimate bandwidth ('scott', 'silverman' or a scalar)
            Defaults to scipy's default for KDE
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        kernel = sts.gaussian_kde(self.PSDsize, bw_method = bw_method, weights = self.PSD)
        x = np.linspace(self.min, self.max, 1000) if self.linear else np.logspace(self.min, self.max, 1000)    
        y = kernel(x) * self.ZeroMoment() * (self.PSDbounds[1] - self.PSDbounds[0])
        
        if fill:
            axes.fill_between(x * scale, y, np.zeros(len(y)), *args, **kwargs)
        else:
            axes.plot(x * scale, y, *args, **kwargs)
        self.setAxes(axes, logX, logY) 
            
    def PlotHistogram(self, axes, outline = 'outline bins', fill = True, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the PSD as a histogram
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        outline : str (optional)
            How to outline the bins ('no outline', 'outline bins', 'outline top')
            Defaults to 'outline bins'
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        if outline == 'outline bins':
            xCoord, yCoord = np.zeros(1 + 3 * self.bins), np.zeros(1 + 3 * self.bins)
            xCoord[0], xCoord[1::3], xCoord[2::3], xCoord[3::3] = self.PSDbounds[0], self.PSDbounds[:-1], self.PSDbounds[1:], self.PSDbounds[1:]
            yCoord[1::3], yCoord[2::3] = self.PSD, self.PSD
        else:
            xCoord, yCoord = np.zeros(1 + 2 * self.bins), np.zeros(1 + 2 * self.bins)
            xCoord[0], xCoord[1::2], xCoord[2::2] = self.PSDbounds[0], self.PSDbounds[:-1], self.PSDbounds[1:]
            yCoord[1::2], yCoord[2::2] = self.PSD, self.PSD

        if outline != 'no outline':
            axes.plot(xCoord * scale, yCoord, *args, **kwargs)
            if fill:
                axes.fill_between(xCoord * scale, yCoord, np.zeros(len(yCoord)), alpha=0.3, *args, **kwargs)
        else:
            axes.fill_between(xCoord * scale, yCoord, np.zeros(len(yCoord)), *args, **kwargs)
        self.setAxes(axes, logX, logY)

    def PlotCDF(self, axes, logX = False, scale = 1, order = 0, *args, **kwargs):
        '''
        Plots cumulative size distribution
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        order : int (optional)
            Moment of specified order
        *args, **kwargs - extra arguments for plotting
        '''
        axes.plot(self.PSDsize * scale, self.CumulativeMoment(order) / self.Moment(order), *args, **kwargs)
        self.setAxes(axes, logX, False) 
        axes.set_ylim([0, 1])
        
    def setAxes(self, axes, logX = False, logY = False): 
        '''
        Sets x- and y-axis to linear or log scale
        
        Parameters
        ----------
        axes : Axis
            Axis to plot on
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        '''    
        if logX:
            if self.min == 0:
                axes.set_xlim([self.PSDbounds[1], self.max])
            else:
                axes.set_xlim([self.min, self.max])
            axes.set_xscale('log')
        else:
            axes.set_xlim([self.min, self.max])

        #Don't set y limits if the PSD is empty
        if any(self.PSD > 0): 
            if logY:
                axes.set_ylim([1e-1, np.amax([1.1 * np.max(self.PSD), 1])])
                axes.set_yscale('log')
            else:
                axes.set_ylim([0, np.amax([1.1 * np.max(self.PSD), 1])])
        
        axes.set_ylabel('Frequency')
            
                    
        
        